ARPC = LibStub("AceAddon-3.0"):NewAddon("AngryRaidPointsClient", "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0", "AceTimer-3.0")
local LootItem = {}
local libS = LibStub("AceSerializer-3.0")
local libC = LibStub("LibCompress")
local libCE = libC:GetAddonEncodeTable()

local default_channel = "RAID"

local ARPC_Version = 'v1.3.4'

BINDING_HEADER_ARPC		= "Angry Raid Points Client"
BINDING_NAME_ARPC_TOGGLE = "Toggle Window"

local MINIMUM_MS = 10
local MINIMUM_OS = 1

local STATUS_NOT_BIDDING = 0
local STATUS_OFF_SPEC = 1
local STATUS_MAIN_SPEC = 2

local STATE_NOT_STARTED = 0
local STATE_LAST_CALL = 1
local STATE_BIDDING_MS = 2
local STATE_BIDDING_OS = 3
local STATE_BIDDING_MSTRIAL = 4
local STATE_BIDDING_OSTRIAL = 5
local STATE_FINISHED = 6

local itemIdBonuses = {
	[120210] = "M", [120284] = "M", [120281] = "H", [120207] = "H", -- Essence of the Iron Protector
	[120209] = "M", [120283] = "M", [120206] = "H", [120280] = "H", -- Essence of the Iron Conqueror
	[120211] = "M", [120285] = "M", [120282] = "H", [120208] = "H", -- Essence of the Iron Vanquisher
}

local protocolVersion = 2 -- ARP protocol 2
local comPrefix = "ARP"
if protocolVersion > 1 then comPrefix = comPrefix .. protocolVersion end

-- Format for our addon communication
--
-- { "LOOT",
--   {
--     { [Item Link], [Quantity], [Handler or nil] },
--     ...
--   }
-- }
-- 
-- { "BIDLIST",
--   [Item Link],
--   {
--    { // Member MS tier
--      { [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--      ...
--    },
--    { // Member OS tier
--      { [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--      ...
--    },
--    { // Trial MS tier
--      { [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--      ...
--    },
--    { // Trial OS tier
--      { [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--      ...
--    },
--   },
--   [State, 0=bidding not started, 1=last call announced, 2=Member MS tier bidding, 3=Member OS tier bidding, 4=Trial MS tier bidding, 5=Trial OS tier bidding],
--   [HighBidder (name) or nil],
--   [CurrentBidder (numeric index) or nil if not started],
--   [Quantity remaining, copied here from LOOT for convenience],
--   [Number of item currently being bid, will be 1 if bidding hasn't started yet],
--   [Minimum bid required for current bidder, or nil if not started]
-- }
--
-- { "POINTS", 
--   [TimeStamp],
--   { [Player]: [Points],
--      ...
--	 }
-- }
-- 
-- { "BIDDING", [Item Link], [Bidding? true=bid, false=remove yourself], [MS? true or false] }
-- 
-- { "BIDDING_ACK", [Item Link], [Error text if there is an error, or nil], [Tier bidding in after BIDDING operation: 1=M MS, 2=M OS, 3=T MS, 4=T OS, or nil if not in any bidding lists] }
-- 
-- { "BID", [Item Link], [Bid Amount, or 0 to pass] }
-- 
-- { "BID_ACK", [Item Link], [Error text if there is an error, or nil] }
-- 
-- { "WINNER", [Item Link], [Number of the one won, 1 if there's only 1], [Player, or nil if DE'ed], [Winning Bid Amount, or nil if DE'ed], [Loot Log Text] }
-- 
-- { "VER_QUERY" }
-- { "VERSION", [Version], [SVN Revision #] }
-- { "ASSIGN", [Item Link], [Quantity] }
-- { "REQUEST_STATUS" }
-- { "QUERY" }
-- { "UPDATE" }
-- { "READY" }
-- { "RESET" }

-- Constants for dealing with our addon communication
local COMMAND = 1

local LOOT_Items = 2

local LOOT_Items_Item_Link = 1
local LOOT_Items_Quantity = 2
local LOOT_Items_Handler = 3

local BIDLIST_Item_Link = 2
local BIDLIST_Bidders = 3
local BIDLIST_State = 4
local BIDLIST_HighBidder = 5
local BIDLIST_CurrentBidder = 6
local BIDLIST_Quantity = 7
local BIDLIST_Number = 8
local BIDLIST_MinBid = 9

local BIDLIST_Bidders_MS = 1
local BIDLIST_Bidders_OS = 2
local BIDLIST_Bidders_MSTrial = 3
local BIDLIST_Bidders_OSTrial = 4

local BIDLIST_Bidders_Player = 1
local BIDLIST_Bidders_Points = 2
local BIDLIST_Bidders_Bid = 3

local BIDDING_Item_Link = 2
local BIDDING_Bidding = 3
local BIDDING_MS = 4

local BIDDING_ACK_Item_Link = 2
local BIDDING_ACK_Error = 3
local BIDDING_ACK_BiddingStatus = 4

local POINTS_TimeStamp = 2
local POINTS_PointsData = 3

local BID_Item_Link = 2
local BID_Bid = 3

local BID_ACK_Item_Link = 2
local BID_ACK_Error = 3

local WINNER_Item_Link = 2
local WINNER_Number = 3
local WINNER_Player = 4
local WINNER_Bid = 5
local WINNER_Loot_Log_Text = 6

local VERSION_Version = 2
local VERSION_SVN_Revision = 3

local ASSIGN_Item_Link = 2
local ASSIGN_Quantity = 3

local lootList = {}

local function Error(message)
	ARPC:Print(RED_FONT_COLOR_CODE.."Error:|r "..message)
end

function ARPC_Toggle()
	ARPC.window:Toggle()
end

local function TestItems()
	ARPC:ProcessMessage("Stingray", {
		"LOOT",
		{
			{ "|cffa335ee|Hitem:96685:0:0:0:0:0:0:0:90:0:0|h[Handwraps of the Witch Doctor]|h|r", 1, nil },
			{ "|cffa335ee|Hitem:96808:0:0:0:0:0:0:0:90:0:0|h[Tortos' Discarded Shell]|h|r", 1, nil }
		}
	})
	ARPC:ProcessMessage("Stingray", {
		"LOOT",
		{
			{ "|cffa335ee|Hitem:96685:0:0:0:0:0:0:0:90:0:0|h[Handwraps of the Witch Doctor]|h|r", 1, nil },
			{ "|cffa335ee|Hitem:96808:0:0:0:0:0:0:0:90:0:0|h[Tortos' Discarded Shell]|h|r", 2, "Ermod" },
			{ "|cffa335ee|Hitem:96487:0:0:0:0:0:0:0:90:0:0|h[Ritual Dagger of the Mind's Eye]|h|r", 3, "Ermod" },
			{ "|cffa335ee|Hitem:96537:0:0:0:0:0:0:0:90:0:0|h[Passionfire Choker]|h|r", 1, "Gulvan" },
			{ "|cffa335ee|Hitem:96507:0:0:0:0:0:0:0:90:0:0|h[Stolen Relic of Zuldazar]|h|r", 2, "Yukio" }
		}
	})

	ARPC:ProcessMessage("Stingray", {
		"BIDLIST",
		"|cffa335ee|Hitem:96808:0:0:0:0:0:0:0:90:0:0|h[Tortos' Discarded Shell]|h|r",
		{
			{ { "Ermod", 234, 12 }, { "Stingray", 342, 0 }, { "Yukio", 1234, 15 }, { "Zanrith", 1434, nil } },
			{ { "Nachwarrior", 1534, nil } },
			{},
			{ { "Endule", 9999, nil } }
		},
		STATE_BIDDING_MS,
		"Yukio",
		4,
		3,
		2,
		16
	})
	ARPC:ProcessMessage("Stingray", { "WINNER", "|cffa335ee|Hitem:96808:0:0:0:0:0:0:0:90:0:0|h[Tortos' Discarded Shell]|h|r", 1, "Yukio", 14 })

	ARPC:ProcessMessage("Stingray", {
		"BIDLIST",
		"|cffa335ee|Hitem:96507:0:0:0:0:0:0:0:90:0:0|h[Stolen Relic of Zuldazar]|h|r",
		{
			{ { "Tisma", 230, nil }, { "Ermod", 234, nil } },
			{ { "Klavak", 534, nil }, { "Nachwarrior", 1534, nil } },
			{},
			{}
		},
		STATE_BIDDING_MS,
		nil,
		2,
		2,
		1,
		10
	})
	ARPC:ProcessMessage("Stingray", {
		"BIDLIST",
		"|cffa335ee|Hitem:96537:0:0:0:0:0:0:0:90:0:0|h[Passionfire Choker]|h|r",
		{ {}, {}, {}, {} },
		STATE_LAST_CALL,
		nil,
		nil,
		nil,
		1,
		1,
		nil
	})
	ARPC:ProcessMessage("Stingray", { "WINNER", "|cffa335ee|Hitem:96507:0:0:0:0:0:0:0:90:0:0|h[Stolen Relic of Zuldazar]|h|r", 1, "Ermod", 11 })
end

local function TestItems2()
	ARPC:ProcessMessage("Stingray", {
		"LOOT",
		{
			{ GetInventoryItemLink("player", GetInventorySlotInfo("MainHandSlot")), 1, nil },
			{ GetInventoryItemLink("player", GetInventorySlotInfo("BackSlot")), 2, "Ermod" },
			{ GetInventoryItemLink("player", GetInventorySlotInfo("Trinket0Slot")), 3, nil },
			{ GetInventoryItemLink("player", GetInventorySlotInfo("NeckSlot")), 1, "Gulvan" }
		}
	})
end

function ARPC:ReceiveMessage(prefix, data, channel, sender)
	if prefix ~= comPrefix then return end
	
	local one = libCE:Decode(data) -- Decode the compressed data
	
	local two, message = libC:Decompress(one) -- Decompress the decoded data
	
	if not two then error("Error decompressing: " .. message); return end
	
	local success, final = libS:Deserialize(two) -- Deserialize the decompressed data
	if not success then error("Error deserializing " .. final); return end

	self:ProcessMessage( sender, final )
end

function ARPC:SendMessage(data, channel, target)
	local one = libS:Serialize( data )
	local two = libC:CompressHuffman(one)
	local final = libCE:Encode(two)
	local destChannel = channel or default_channel

	if destChannel ~= "RAID" or IsInRaid(LE_PARTY_CATEGORY_HOME) then
		self:SendCommMessage(comPrefix, final, destChannel, target, "NORMAL")
	end
end

function ARPC:ProcessMessage(sender, data)
	-- ARPC:Print("Got "..data[COMMAND].." from "..sender)
	if data[COMMAND] == "LOOT" then
		local count = 0
		local links = {}
		for index, item in pairs( data[LOOT_Items] ) do
			local result = LootItem:add(item)
			if result then count = count + 1 end
			links[ item[LOOT_Items_Item_Link] ] = true
		end
		for link,item in pairs( lootList ) do
			if not links[ link ] then
				item:ItemFinished()
			end
		end
		if count > 0 then
			if not ARPC.window.frame:IsShown() then ARPC.window.frame:Show() end
			LootItem:ClearFinished()
		end
	elseif data[COMMAND] == "BIDLIST" then
		if LootItem:get( data[BIDLIST_Item_Link] ) then
			LootItem:get(data[BIDLIST_Item_Link]):SetBidList(data)
		end
	elseif data[COMMAND] == "BIDDING_ACK" then
		if LootItem:get( data[BIDDING_ACK_Item_Link] ) then
			LootItem:get( data[BIDDING_ACK_Item_Link] ):ReceivedBiddingAck( data[BIDDING_ACK_BiddingStatus], data[BIDDING_ACK_Error] )
		end
	elseif data[COMMAND] == "BID_ACK" then
		if LootItem:get( data[BID_ACK_Item_Link] ) then
			LootItem:get( data[BID_ACK_Item_Link] ):ReceivedBidAck( data[BID_ACK_Error] )
		end
	elseif data[COMMAND] == "WINNER" then
		if LootItem:get( data[WINNER_Item_Link] ) then
			LootItem:get( data[WINNER_Item_Link] ):ProcessWinner(data[WINNER_Number], data[WINNER_Player], data[WINNER_Bid])
		end
	elseif data[COMMAND] == "POINTS" then
		ARPC:ProcessPoints( data[POINTS_PointsData], data[POINTS_TimeStamp] )
	end
end

local function HasARP()
	return not not SlashCmdList.ACECONSOLE_ARP
end

local waitingForItemData = {}
function ARPC:GET_ITEM_INFO_RECEIVED()
	if #waitingForItemData < 1 then return end
	local newWaiting = {}
	for index, item_link in pairs(waitingForItemData) do
		if GetItemInfo(item_link) then
			local loot = lootList[item_link]
			if loot then
				loot:UpdateFrameInitial()
			end
		else
			table.insert(newWaiting, item_link)
		end
	end
	waitingForItemData = newWaiting
end

local function UpdateStatusBar()
	if ARPC_CurrentPoints == nil then
		ARPC.window.statustext:SetText("Current Points: Not Set")
	else
		ARPC.window.statustext:SetText("Current Points: "..ARPC_CurrentPoints)
	end
end

function ARPC:ProcessPoints(data, timestamp)
	local new_points = data[UnitName("player")]
	if new_points then
		ARPC_CurrentPoints = new_points
		UpdateStatusBar()
	end
end

local function CreateWindow()
	ARPC.window = ARPC_CreateWindow()
	UpdateStatusBar()
end

local function CreateItemButton(parent, status, icon, text, ...)
	local f = CreateFrame("Button", nil, parent)
	f:SetPoint(...)
	f:SetWidth(24)
	f:SetHeight(24)
	f:SetNormalTexture("Interface\\BUTTONS\\UI-GroupLoot-"..icon.."-Up")
	f:SetPushedTexture("Interface\\BUTTONS\\UI-GroupLoot-"..icon.."-Down")
	f:SetHighlightTexture("Interface\\BUTTONS\\UI-GroupLoot-"..icon.."-Highlight", "ADD")
	
	f:SetScript("OnEnter", function(frame)
		GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
		GameTooltip:SetText(text)
		GameTooltip:Show()
	end)
	f:SetScript("OnLeave", function(frame) GameTooltip:Hide() end)
	
	f:SetScript("OnDisable", function(frame) frame:SetAlpha(0.3) end)
	f:SetScript("OnEnable", function(frame) frame:SetAlpha(1) end)

	f:SetScript("OnClick", function(frame) frame:GetParent().obj:SetStatus(status) end)

	return f
end

local widgetReusePool = {}
local frameCount = 0
function LootItem:CreateFrame()
	if #widgetReusePool > 0 then
		local group = widgetReusePool[#widgetReusePool]
		table.remove(widgetReusePool, #widgetReusePool)
		group.obj = self
		self.group = group
		self:UpdateFrameInitial()
		self:UpdateFrame()
		self:UpdateBiddersList()
		ARPC.window:AddChild(group)
		return 
	end
	frameCount = frameCount + 1

	local group = CreateFrame("Frame")
	group.obj = self
	group:SetHeight(52)
	group:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 3, right = 3, top = 5, bottom = 3 }
	})
	group:SetBackdropColor(0.1,0.1,0.1,0.5)
	group:SetBackdropBorderColor(0.4,0.4,0.4)

	local icon = CreateFrame("Button", nil, group)
	group.icon = icon
	icon:SetScript("OnEnter", function(frame)
		GameTooltip:SetOwner(frame, "ANCHOR_LEFT", -4, -32)
		GameTooltip:SetHyperlink(frame:GetParent().obj.item_link)
		GameTooltip:Show()
	end)
	icon:SetScript("OnClick", function(frame)
		if IsModifiedClick() then
			HandleModifiedItemClick(frame:GetParent().obj.item_link)
		end
	end)
	icon:SetScript("OnLeave", function(frame) GameTooltip:Hide() end)
	icon:SetWidth(32)
	icon:SetHeight(32)
	icon:SetPoint("TOPLEFT", group, "TOPLEFT", 10, -10)

	local item_lbl = group:CreateFontString()
	group.item_lbl = item_lbl
	item_lbl:SetFontObject("GameFontNormalLarge")
	item_lbl:SetJustifyH("LEFT")
	item_lbl:SetPoint("TOPLEFT", group, "TOPLEFT", 50, -11)
	item_lbl:SetPoint("TOPRIGHT", group, "TOPRIGHT", -104, -11)

	local status_lbl = group:CreateFontString()
	status_lbl:SetFontObject("GameFontHighlight")
	status_lbl:SetPoint("TOPLEFT", item_lbl, "BOTTOMLEFT", 0, -6)
	status_lbl:SetPoint("TOPRIGHT", item_lbl, "BOTTOMRIGHT", 0, -6)
	status_lbl:SetJustifyH("LEFT")
	group.status_lbl = status_lbl
	
	local fontName, fontHeight, fontFlags = item_lbl:GetFont()
	local _, fontHeight2, _ = status_lbl:GetFont()
	item_lbl:SetFont(fontName, (fontHeight + fontHeight2) / 2, fontFlags)

	local bidders_lbl = group:CreateFontString()
	bidders_lbl:SetFontObject("GameFontHighlight")
	bidders_lbl:SetPoint("TOPLEFT", status_lbl, "BOTTOMLEFT", 0, -4)
	bidders_lbl:SetPoint("TOPRIGHT", status_lbl, "BOTTOMRIGHT", -90, -4)
	bidders_lbl:SetJustifyH("LEFT")
	group.bidders_lbl = bidders_lbl

	local bidders2_lbl = group:CreateFontString()
	bidders2_lbl:SetFontObject("GameFontHighlight")
	bidders2_lbl:SetPoint("TOPLEFT", status_lbl, "BOTTOMRIGHT", -95, -4)
	bidders2_lbl:SetPoint("TOPRIGHT", status_lbl, "BOTTOMRIGHT", 0, -4)
	bidders2_lbl:SetJustifyH("RIGHT")
	group.bidders2_lbl = bidders2_lbl

	group.ms_btn = CreateItemButton(group, STATUS_MAIN_SPEC, "Dice", "Bidding Main Spec", "TOPRIGHT", -64, -14)
	group.os_btn = CreateItemButton(group, STATUS_OFF_SPEC, "Coin", "Bidding Off Spec", "TOPRIGHT", -36, -14)
	group.pass_btn = CreateItemButton(group, STATUS_NOT_BIDDING, "Pass", "Not Bidding", "TOPRIGHT", -8, -12)

	if HasARP() then
		local startbid = CreateFrame("Button", nil, group, "UIPanelButtonTemplate")
		startbid:SetPoint("TOPRIGHT", group, "TOPRIGHT", -96, -24)
		startbid:SetWidth(56)
		local startbidtext = startbid:GetFontString()
		startbidtext:ClearAllPoints()
		startbidtext:SetPoint("TOPLEFT", 10, -1)
		startbidtext:SetPoint("BOTTOMRIGHT", -10, 1)
		startbidtext:SetJustifyV("MIDDLE")
		startbidtext:SetText("Start")
		startbid:SetScript("OnClick", function(frame) SlashCmdList.ACECONSOLE_ARP("bid", nil) end)
		startbid:Hide()
		group.startbid = startbid

		local reminder = CreateFrame("Button", nil, group, "UIPanelButtonTemplate")
		reminder:SetPoint("TOPRIGHT", group, "TOPRIGHT", -96, -24)
		reminder:SetWidth(70)
		local remindertext = reminder:GetFontString()
		remindertext:ClearAllPoints()
		remindertext:SetPoint("TOPLEFT", 10, -1)
		remindertext:SetPoint("BOTTOMRIGHT", -10, 1)
		remindertext:SetJustifyV("MIDDLE")
		remindertext:SetText("Remind")
		reminder:SetScript("OnClick", function(frame) SlashCmdList.ACECONSOLE_ARP("reminder", nil) end)
		reminder:Hide()
		group.reminder = reminder

		local lastcall = CreateFrame("Button", nil, group, "UIPanelButtonTemplate")
		lastcall:SetPoint("TOPRIGHT", startbid, "TOPLEFT", -2, 0)
		lastcall:SetWidth(74)
		local lastcalltext = lastcall:GetFontString()
		lastcalltext:ClearAllPoints()
		lastcalltext:SetPoint("TOPLEFT", 10, -1)
		lastcalltext:SetPoint("BOTTOMRIGHT", -10, 1)
		lastcalltext:SetJustifyV("MIDDLE")
		lastcalltext:SetText("Last Call")
		lastcall:SetScript("OnClick", function(frame) SlashCmdList.ACECONSOLE_ARP("lastcall", nil) end)
		lastcall:Hide()
		group.lastcall = lastcall
	end


	local bid = CreateFrame("Button", nil, group, "UIPanelButtonTemplate")
	bid:SetPoint("TOPRIGHT", group, "TOPRIGHT", -96, -24)
	bid:SetWidth(40)
	local bidtext = bid:GetFontString()
	bidtext:ClearAllPoints()
	bidtext:SetPoint("TOPLEFT", 10, -1)
	bidtext:SetPoint("BOTTOMRIGHT", -10, 1)
	bidtext:SetJustifyV("MIDDLE")
	bidtext:SetText("Bid")
	bid:SetScript("OnClick", function(frame) frame:GetParent().obj:ShowBidPopup() end)
	bid:Hide()
	group.bid = bid

	self.group = group
	self:UpdateFrameInitial()
	self:UpdateFrame()
	ARPC.window:AddChild(group)
end

local function UnitString(unit)
	if unit then
		local _, classFileName = UnitClass(unit)
		if classFileName then
			return "|c"..(RAID_CLASS_COLORS[classFileName].colorStr)..unit
		else
			return unit
		end
	else
		return ""
	end
end

function ARPC:UpdateAllVisibilities()
	for item_link, loot in pairs(lootList) do
		loot:UpdateVisiblity()
	end
end

function LootItem:UpdateVisiblity()
	self:UpdateBiddersList()
end

function LootItem:UpdateFrameInitial()
	local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(self.item_link)

	if link then
		local quantity = self.quantity
		if self.bidList and self.bidList[BIDLIST_Quantity] then quantity = self.bidList[BIDLIST_Quantity] + self.bidList[BIDLIST_Number] - 1 end
		
		local quantity_txt
		if  quantity > 1 then
			quantity_txt = "x"..quantity
		else
			quantity_txt = ""
		end

		local desc_txt = {}
		local itemString = { strsplit(":", strmatch(link, '%|H(item:[0-9:]+)%|h')) }
		local numBonus = tonumber(itemString[13])
		for i = 1, numBonus do
			local bonusId = tonumber(itemString[13 + i])
			if bonusId == 560 or bonusId == 561 or bonusId == 562 then
				table.insert(desc_txt, "WF")
			elseif bonusId == 566 or bonusId == 570 then
				table.insert(desc_txt, 1, "H")
			elseif bonusId == 567 or bonusId == 569 then
				table.insert(desc_txt, 1, "M")
			elseif bonusId == 563 or bonusId == 564 or bonusId == 565 then
				table.insert(desc_txt, "S")
			elseif bonusId == 40 then
				table.insert(desc_txt, "A")
			elseif bonusId == 41 then
				table.insert(desc_txt, "L")
			elseif bonusId == 42 then
				table.insert(desc_txt, "SP")
			elseif bonusId == 43 then
				table.insert(desc_txt, "I")
			end
		end
		local itemId = tonumber(itemString[2])
		if itemIdBonuses[itemId] then
			table.insert(desc_txt, 1, itemIdBonuses[itemId])
		end

		if #desc_txt > 0 then
			self.group.item_lbl:SetText(link..quantity_txt..GREEN_FONT_COLOR_CODE.." "..table.concat(desc_txt, " "))
		else
			self.group.item_lbl:SetText(link..quantity_txt)
		end
		self.group.icon:SetNormalTexture(texture)
	else
		table.insert(waitingForItemData, self.item_link)
		self.group.item_lbl:SetText(RED_FONT_COLOR_CODE.."Fetching item info...")
		self.group.icon:SetNormalTexture("Interface\\Icons\\INV_Misc_QuestionMark")
	end
end


function LootItem:UpdateFrame()
	local winners = ""
	for index,list in pairs(self.winners) do
		if index > 1 then winners = winners .. ", " end
		if list[1] then
			winners = winners .. UnitString(list[1]).." "..list[2].."|r"
		else
			winners = winners .. GRAY_FONT_COLOR_CODE .. "Disenchanted|r"
		end
	end

	if not self.bidList then
		if self.handler then
			self.group.status_lbl:SetText("Bidding Open - "..UnitString(self.handler))
		else
			self.group.status_lbl:SetText("")
		end
	elseif self.bidList[BIDLIST_State] == STATE_NOT_STARTED and self.handler then
		self.group.status_lbl:SetText("Bidding Open - "..UnitString(self.handler))
	elseif self.bidList[BIDLIST_State] == STATE_LAST_CALL and self.handler then
		self.group.status_lbl:SetText("Last Call - "..UnitString(self.handler))
	elseif self.bidList[BIDLIST_State] == STATE_FINISHED then
		self.group.status_lbl:SetText(GREEN_FONT_COLOR_CODE .. "Finished|r - " .. winners)
	elseif self.bidList[BIDLIST_State] >= STATE_BIDDING_MS and self.bidList[BIDLIST_State] < STATE_FINISHED then
		local status = nil
		local bidder = self.bidList[BIDLIST_HighBidder]
		local bid = self:GetHighBidAmount()
		if bidder and bid then
			status = UnitString(bidder) .. " " .. bid .. "|r"
		else
			status = "No bids yet"
		end
		if self:CurrentBidder() then
			status = status .. " - " .. UnitString(self:CurrentBidder()) .. "'s bid|r"
		end
		if winners ~= "" then
			status = status .. " - " .. winners
		end
		
		self.group.status_lbl:SetText(status)
	else
		self.group.status_lbl:SetText("")
	end
	
	if HasARP() then
		if self.handler == UnitName('player') and self:Biddingable() then
			self.group.startbid:Show()
			self.group.lastcall:Show()
			self.group.reminder:Hide()
			self.group.bid:Hide()
		elseif self:CurrentBidder() == UnitName('player') then
			self.group.startbid:Hide()
			self.group.lastcall:Hide()
			self.group.reminder:Hide()
			self.group.bid:Show()
		elseif self.handler == UnitName('player') and self:Bidable() then
			self.group.startbid:Hide()
			self.group.lastcall:Hide()
			self.group.reminder:Show()
			self.group.bid:Hide()
		else
			self.group.startbid:Hide()
			self.group.lastcall:Hide()
			self.group.reminder:Hide()
			self.group.bid:Hide()
		end
	else
		if self:CurrentBidder() == UnitName('player') then
			self.group.bid:Show()
		else
			self.group.bid:Hide()
		end
	end

	if self:Biddingable() then -- Able to be bid on
		self.group.ms_btn:Enable()
		self.group.os_btn:Enable()
		self.group.pass_btn:Enable()
	else
		self.group.ms_btn:Disable()
		self.group.os_btn:Disable()
		self.group.pass_btn:Disable()
	end


	if self.status == STATUS_MAIN_SPEC then
		self.group.ms_btn:GetNormalTexture():SetVertexColor(0.616, 0.149, 0.114)
		self.group.os_btn:GetNormalTexture():SetVertexColor(1, 1, 1)
		self.group.pass_btn:GetNormalTexture():SetVertexColor(1, 1, 1)
	elseif self.status == STATUS_OFF_SPEC then
		self.group.ms_btn:GetNormalTexture():SetVertexColor(1, 1, 1)
		self.group.os_btn:GetNormalTexture():SetVertexColor(0.616, 0.149, 0.114)
		self.group.pass_btn:GetNormalTexture():SetVertexColor(1, 1, 1)
	elseif self.status == STATUS_NOT_BIDDING then
		self.group.ms_btn:GetNormalTexture():SetVertexColor(1, 1, 1)
		self.group.os_btn:GetNormalTexture():SetVertexColor(1, 1, 1)
		self.group.pass_btn:GetNormalTexture():SetVertexColor(0.616, 0.149, 0.114)
	else
		self.group.ms_btn:GetNormalTexture():SetVertexColor(1, 1, 1)
		self.group.os_btn:GetNormalTexture():SetVertexColor(1, 1, 1)
		self.group.pass_btn:GetNormalTexture():SetVertexColor(1, 1, 1)
	end
end

function LootItem:UpdateBiddersList()
	if not self.bidList or self.bidList[BIDLIST_State] == STATE_FINISHED then
		self.group.bidders_lbl:SetText("")
		self.group.bidders2_lbl:SetText("")
		self.group:SetHeight(52)
		ARPC.window:DoLayout()
		return
	end
	local text = ""
	local text_right = ""
	local height = 0
	local first_tier = nil
	if not (ARPC_Collapse and self.status ~= STATUS_MAIN_SPEC and self.status ~= STATUS_OFF_SPEC) then
		for index,list in pairs(self.bidList[BIDLIST_Bidders]) do
			for offset,bidder in pairs(list) do
				if text ~= "" then text = text .. "\n" end
				if text_right ~= "" then text_right = text_right .. "\n" end
				if not first_tier then first_tier = index end
				
				local current_tier
				if self.bidList[BIDLIST_State] < STATE_BIDDING_MS then
					current_tier = index == first_tier
				else
					current_tier = (self.bidList[BIDLIST_State] - STATE_BIDDING_MS + 1) == index
				end
				
				local tier_color = nil

				if current_tier then
					text = text .. UnitString(bidder[BIDLIST_Bidders_Player]) .. "|r"	
				else
					text = text .. GRAY_FONT_COLOR_CODE .. (bidder[BIDLIST_Bidders_Player]) .. "|r"
					tier_color = GRAY_FONT_COLOR_CODE
				end
				
				local bid = bidder[BIDLIST_Bidders_Bid]
				if bid == 0 then
					text = text .. (tier_color or "") .. " - " .. (tier_color or YELLOW_FONT_COLOR_CODE) .. "Pass|r"
				elseif bid then
					if bidder[BIDLIST_Bidders_Player] == self.bidList[BIDLIST_HighBidder] then
						text = text .. (tier_color or "") .. " - " .. (tier_color or GREEN_FONT_COLOR_CODE) .. bid .. "|r"
					else
						text = text .. (tier_color or "") .. " - " .. (tier_color or YELLOW_FONT_COLOR_CODE) .. bid .. "|r"
					end
				end
				
				local group = ""
				if index == BIDLIST_Bidders_OS then
					group = "OS "
				elseif index == BIDLIST_Bidders_MSTrial then
					group = "Trial "
				elseif index == BIDLIST_Bidders_OSTrial then
					group = "Trial OS "
				end
				
				text_right = text_right .. (tier_color or "") .. group .. bidder[BIDLIST_Bidders_Points] .. "|r"
				
				height = height + 13
			end
		end
	end
	self.group.bidders_lbl:SetText(text)
	self.group.bidders2_lbl:SetText(text_right)
	self.group:SetHeight(52 + self.group.bidders_lbl:GetHeight())
	ARPC.window:DoLayout()
end

function LootItem:ItemFinished()
	if self.bidList then
		self.bidList[BIDLIST_State] = STATE_FINISHED
		self:UpdateFrame()
		self:UpdateBiddersList()
	else
		self:RemoveItem()
	end
end

function LootItem:RemoveItem()
	ARPC.window:RemoveChild(self.group)
	
	self.group.obj = nil
	tinsert(widgetReusePool, self.group)
	self.group = nil
	
	lootList[ self.item_link ] = nil
end

function LootItem:ClearFinished()
	for link,item in pairs( lootList ) do
		if item.bidList and item.bidList[BIDLIST_State] == STATE_FINISHED then
			item:RemoveItem()
		end
	end
end

function LootItem:Error(text)
	Error(YELLOW_FONT_COLOR_CODE.."("..self.item_link..YELLOW_FONT_COLOR_CODE..")|r "..text)
end
function LootItem:Alert(text)
	ARPC:Print("("..self.item_link..")|r "..text)
end

function LootItem:CurrentMinimumBid()
	return self.bidList[BIDLIST_MinBid]
end

function LootItem:CurrentBidder()
	if not self.bidList or not self.bidList[BIDLIST_CurrentBidder] then return nil end
	local list = nil
	if self.bidList[BIDLIST_State] == STATE_BIDDING_MS then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_MS]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_OS then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_OS]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_MSTRIAL then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_MSTrial]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_OSTRIAL then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_OSTrial]
	end
	if not list then return nil end

	return list[ self.bidList[BIDLIST_CurrentBidder] ][ BIDLIST_Bidders_Player ]
end

function LootItem:ShowBidPopup()
	local popup_name = "ARP_BID_POPUP_"..self.item_link
	local item_link = self.item_link

	if StaticPopupDialogs[popup_name] == nil then

		StaticPopupDialogs[popup_name] = {
			button1 = "Send",
			button2 = "Dismiss",
			button3 = "Pass",
			OnAccept = function(self)
					local text = self.editBox:GetText()
					return not LootItem:get(item_link):EnteredBid(text)
			end,
			EditBoxOnEnterPressed = function(self)
					local text = self:GetParent().editBox:GetText()
					if LootItem:get(item_link):EnteredBid(text) then
						self:GetParent():Hide()
					end
			end,
			OnAlt = function(self)
					local text = self.editBox:GetText()
					return not LootItem:get(item_link):EnteredBid('pass')
			end,
			hasEditBox = true,
			whileDead = true,
			EditBoxOnEscapePressed = function(self) self:GetParent():Hide() end,
			hideOnEscape = 1,
			preferredIndex = 3
		}
	end
	
	local text = "It is your turn to bid on "..item_link
	local highBidAmt = self:GetHighBidAmount()
	if highBidAmt then
		text = text .. "\n\nThe highest bidder is "..UnitString(self.bidList[BIDLIST_HighBidder]).."|r for "..YELLOW_FONT_COLOR_CODE..highBidAmt..".|r"
		text = text .. " Your minimum bid is "..YELLOW_FONT_COLOR_CODE..self:CurrentMinimumBid().."|r."
	else
		text = text .. "\n\nThere have been no bids yet. The minimum bid is "..YELLOW_FONT_COLOR_CODE..self:CurrentMinimumBid().."|r."
	end
	StaticPopupDialogs[popup_name].text = text


	for i = 1, STATICPOPUP_NUMDIALOGS do
		local frame = _G['StaticPopup'..i]
		if frame:IsShown() and frame.which == popup_name then
			return
		end
	end
	StaticPopup_Show(popup_name)
end

function LootItem:EnteredBid(text)
	local bid
	local minimum_bid = self:CurrentMinimumBid()

	text = strtrim(strlower(text))
	if text == "yup" then text = "+1" end
	if text == "nope" then text = "pass" end
	if text ~= 'pass' and not strmatch(text, '^\+?%d+$') then
		Error("You have entered an invalid bid. (" .. text .. ")")
		return false
	elseif text == "pass" then
		self:SendBid(0)
		return true
	elseif strsub(text, 1, 1) == "+" then
		bid = minimum_bid - 1 + tonumber(strsub(text, 2))
	else
		bid = tonumber(text)
	end
	if bid < minimum_bid then
		Error("Your bid of "..bid.." is less the minimum bid of "..minimum_bid..".")
		return false
	elseif ARPC_CurrentPoints and bid ~= minimum_bid and bid > ARPC_CurrentPoints then
		Error("Your bid of "..bid.." exceeds your maximum point total.")
		return false
	else
		self:SendBid(bid)
	end
	return true
end

function LootItem:get(item_link)
	return lootList[ item_link ]
end

function LootItem:add(data)
	local loot = lootList[ data[LOOT_Items_Item_Link] ]
	if loot then
		if loot.bidList and loot.bidList[BIDLIST_State] == STATE_FINISHED then
			loot:RemoveItem()
		else
			loot.quantity = data[LOOT_Items_Quantity]
			if loot.handler ~= data[LOOT_Items_Handler] then
				loot.handler = data[LOOT_Items_Handler]
				loot.bidding_sent = false
				loot.bidList = nil
			end
		
			loot:UpdateFrameInitial()
			loot:UpdateFrame()
			loot:UpdateBiddersList()
			loot:CheckBidding()
			return false
		end
	end
	loot = {
		item_link = data[LOOT_Items_Item_Link],
		quantity = data[LOOT_Items_Quantity],
		handler = data[LOOT_Items_Handler],
		bidList = nil,
		
		status = nil, -- Status of user bidding
		bidding_sent = false, -- nil for not sent, false for send but no ack, true for sent and acked
		winners = {}
	}
	lootList[ loot.item_link ] = loot
	setmetatable(loot, self)
	self.__index = self
	loot:CreateFrame()
	return true
end

function LootItem:GetHighBidAmount()
	local list = nil
	if self.bidList[BIDLIST_State] == STATE_BIDDING_MS then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_MS]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_OS then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_OS]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_MSTRIAL then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_MSTrial]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_OSTRIAL then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_OSTrial]
	end
	if list == nil then return nil end

	local highBidder = nil

	for index, entry in pairs( list ) do
		if entry[BIDLIST_Bidders_Bid] and entry[BIDLIST_Bidders_Bid] > 0 and (highBidder == nil or entry[BIDLIST_Bidders_Bid] >= highBidder) then
			highBidder = entry[BIDLIST_Bidders_Bid]
		end
	end

	return highBidder
end

function LootItem:OnCurrentBidder()
	self:ShowBidPopup()
end

function LootItem:Biddingable()
	return not self.bidList or self.bidList[BIDLIST_State] < STATE_BIDDING_MS
end

function LootItem:Bidable()
	return self.bidList[BIDLIST_State] >= STATE_BIDDING_MS and self.bidList[BIDLIST_State] <= STATE_BIDDING_OSTRIAL
end

function LootItem:SetBidList(data)
	self.bidList = data
	self:UpdateFrame()
	self:UpdateBiddersList()
	self:CheckBidding()
	if self:CurrentBidder() == UnitName('player') then
		self:OnCurrentBidder()
	end
end

function LootItem:CheckBidding()
	if self.handler and (not self.bidList or self.bidList[BIDLIST_State] < STATE_BIDDING_MS) and not self.bidding_sent and (self.status == STATUS_OFF_SPEC or self.status == STATUS_MAIN_SPEC) then
		self:SendBidding()
	end
end

function LootItem:SendBid(amount)
	if self.handler and self.bidList and self:CurrentBidder() == UnitName("player") then
		ARPC:SendMessage({ [COMMAND]="BID", [BID_Item_Link]=self.item_link, [BID_Bid]=amount }, "WHISPER", self.handler)
	else 
		self:Error("Unable to bid for item")
	end
end

function LootItem:ReceivedBidAck(error)
	if error then
		self:Error("Bid error: "..error)
	end
end

function LootItem:SendBidding()
	if self.handler and (not self.bidList or self.bidList[BIDLIST_State] < STATE_BIDDING_MS) then
		if self.status == STATUS_NOT_BIDDING then
			if self.bidding_sent then
				ARPC:SendMessage({ [COMMAND]="BIDDING", [BIDDING_Item_Link]=self.item_link, [BIDDING_Bidding]=false }, "WHISPER", self.handler)
			end
		elseif self.status then 
			ARPC:SendMessage({ [COMMAND]="BIDDING", [BIDDING_Item_Link]=self.item_link, [BIDDING_Bidding]=true, [BIDDING_MS]=(self.status == STATUS_MAIN_SPEC) }, "WHISPER", self.handler)
		end			
		self.bidding_sent = true
	else
		self:Error("Couldn't send BIDDING")
	end
end

function LootItem:ReceivedBiddingAck(status, error)
	if error then
		self:Error("Bidding error: "..error)
	end
	if status == BIDLIST_Bidders_MS or status == BIDLIST_Bidders_MSTrial then
		self.status = STATUS_MAIN_SPEC
	elseif  status == BIDLIST_Bidders_OS or status == BIDLIST_Bidders_OSTrial then
		self.status = STATUS_OFF_SPEC
	else
		self.status = STATUS_NOT_BIDDING
	end
	self.bidding_sent = true
	self:UpdateFrame()
end


function LootItem:SetStatus(status)
	if not self:Biddingable() then
		self:Error("Bidding for this item has already started")
		return
	end
	if self.status == status then return end -- Bidding interest is already set to current value

	self.status = status
	self:UpdateFrame()
	self:UpdateVisiblity()
	if self.handler == nil then
		-- ARPC:Print("Bidding for this item hasn't started yet. Your interest will be sent when it has.")
	else
		self:SendBidding()
	end
end

function LootItem:ProcessWinner(number, player, bid)
	self.winners[number] = { player, bid }
	self:UpdateFrame()
	self:UpdateBiddersList()
end

function ARPC:OnInitialize()
	if ARPC_Scale == nil then ARPC_Scale = 1 end
	if ARPC_Collapse == nil then ARPC_Collapse = false end
	
	local options = {
		name = "ARPC",
		handler = ARPC,
		type = "group",
		args = {
			help = {
				type = "execute",
				order = 99,
				name = "Help",
				hidden = true,
				func = function()
					LibStub("AceConfigCmd-3.0").HandleCommand(self, "arpc", "Angry Raid Points Client", "")
				end
			},
			toggle = {
				type = "execute",
				name = "Toggle Window",
				desc = "Shows/hides the main window (also available in game keybindings)",
				func = function() ARPC_Toggle() end
			},
			test = {
				type = "execute",
				name = "Test Items",
				desc = "Loads some test items in",
				func = function() TestItems() end
			},
			test2 = {
				type = "execute",
				name = "Test Items",
				desc = "Loads some test items in",
				func = function() TestItems2() end
			},
			scale = {
				type = "range",
				name = "Scale",
				desc = function() 
					return "Sets the scale of the ARP windows (current:|cFFFF0000 "..ARPC_Scale.."|r)" 
				end,
				min = 0.3,
				max = 3,
				get = function(info) return ARPC_Scale end,
				set = function(info, val)
					ARPC.window:SetScale(val)
				end
			}
		}
	}

	self:RegisterChatCommand("arpc", "ChatCommand")
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Angry Raid Points Client", options)
	
	CreateWindow()
end

function ARPC:ChatCommand(input)
	if not input or input:trim() == "" then
		ARPC_Toggle()
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(self, "arpc", "Angry Raid Points Client", input)
	end
end

function ARPC:OnEnable()
	ARPC:RegisterEvent("GET_ITEM_INFO_RECEIVED")

	ARPC:RegisterComm(comPrefix, "ReceiveMessage")
	
	ARPC:ScheduleTimer("AfterEnable", 5)
end

function ARPC:AfterEnable()
	ARPC:SendMessage({ "REQUEST_STATUS" })
end
